# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import struct, math, random
import py2play.action as action, py2play.character as py2play_character
import soya, soya.soya3d as soya3d, soya.cursor as cursor, soya.game.level as game_level, soya.model as model, soya.ray as ray, soya.particle as particle, soya.widget as widget
    
from soya.math3d import Point, Vector
import slune.globdef as globdef, slune.sound as sound

ACTION_WAIT      = action.ACTION_WAIT
ACTION_JUMP      = "j"
ACTION_TURBO     = "t"
ACTION_ROTUP     = "u"
ACTION_ROTDOWN   = "d"
ACTION_ROTLAT    = "l"
ACTION_DEBUG     = ""
ACTION_STARTGAME = "!"

class Action:
  def __init__(self, action = ACTION_WAIT, mouse_x = 0.0, mouse_y = 0.0, round = None):
    self.round   = round
    self.action  = action
    self.mouse_x = mouse_x
    self.mouse_y = mouse_y
    
  def __repr__(self):
    return "<Action %s at round %s, (%s, %s)>" % (self.action, self.round, self.mouse_x, self.mouse_y)
    
def WRITE_ACTION(file, action):
  file.write(struct.pack("cddi", action.action, action.mouse_x, action.mouse_y, action.round))
  
def READ_ACTION(file):
  return Action(*struct.unpack("cddi", file.read(24)))

action.Action       = Action
action.READ_ACTION  = READ_ACTION
action.WRITE_ACTION = WRITE_ACTION

RIGHT  = Vector(None, 1.0,  0.0,  0.0)
DOWN   = Vector(None, 0.0, -1.0,  0.0)
UP     = Vector(None, 0.0,  1.0,  0.0)
FRONT  = Vector(None, 0.0,  0.0, -1.0)

# Re-useable Points and Vectors

_P  = Point()
_P2 = Point()
_P3 = Point()
_V  = Vector()
_V2 = Vector()

class Pusher: pass
class Competitor: pass

class Character(soya3d.World, Pusher):
  def __init__(self, parent = None):
    soya3d.World.__init__(self, parent)
    
    self.radius      = 0.7
    self.floating    = 0.3
    self.weight      = 1.0
    self.rotup_speed = 7.0
    self.inertia     = 0.9
    
    self.speed  = Vector(self)
    self.inercy = Vector(None)
    self.force  = Vector(None)
    self.solid     = 0
    
    self.camera_target = self #Point(self, 0.0, 1.0, 0.0)
    self.old_speed_x = 0.0
    
    self.CENTER = Point (self, 0.0, 0.3,  0.0)
    #self.UP     = Vector(self, 0.0, 1.0,  0.0)
    #self.RIGHT  = Vector(self, 1.0, 0.0,  0.0)
    #self.FRONT  = Vector(self, 0.0, 0.0, -1.0)
    
    self.scramble = 0.0
    self.next_pos = Point(self, 0.0, 0.0, 0.0)
    
    self.internal = soya3d.World(self)
    self.perso    = soya3d.Volume(self.internal)
    
    self.perso_name = None
    self.vehicle = 1
    self.life = 1.0
    
  def set_perso(self, perso_name):
    self.perso_name = perso_name
    self.perso.set_shape(model.Shape.get(self.perso_name + "1"))

  def displayname(self): return self.perso_name.title()
  
  def set_vehicle(self, vehicle, add_ray = 1):
    self.vehicle = vehicle
    self.internal.set_shape(model.Shape.get("vehicle" + str(self.vehicle)))
    if hasattr(self, "ray"): self.internal.remove(self.ray)
    
    if vehicle == 1:
      self.radius      = 0.7
      self.floating    = 0.3
      self.weight      = 1.0
      self.maxspeedx   = 1.0
      self.maxspeedz   = 1.0
      self.rotup_speed = 7.0
      self.inertia     = 0.8
      
      self.perso.set_xyz(0.0, 0.1, 0.03)
      self.perso_y = 0.1
      
      if add_ray:
        self.ray = ray.Ray(self.internal, RAY_MATERIAL)
        self.ray.set_xyz(-0.3, 0.15, 0.5)
        self.ray.endpoint.set_xyz(0.6, 0.0, 0.0)
        self.ray.length = 20
        
    elif vehicle == 2:
      self.radius      = 0.8
      self.floating    = 0.3
      self.weight      = 1.6
      self.maxspeedx   = 0.9
      self.maxspeedz   = 0.95
      self.rotup_speed = 6.0
      self.inertia     = 0.87
      
      self.perso.set_xyz(0.0, 0.4, -0.2)
      self.perso_y = 0.4
      
      if add_ray:
        self.ray = ray.Ray(self.internal, RAY_MATERIAL2)
        self.ray.set_xyz(-0.5, 0.2, 1.0)
        self.ray.endpoint.set_xyz(1.0, 0.0, 0.0)
        self.ray.length = 20
        
    elif vehicle == 3:
      self.radius      = 0.7
      self.floating    = 0.3
      self.weight      = 0.6
      self.maxspeedx   = 1.3
      self.maxspeedz   = 1.0
      self.rotup_speed = 10.0
      self.inertia     = 0.6
      
      self.perso.set_xyz(0.0, 0.2, 0.35)
      self.perso_y = 0.2
      
      if add_ray:
        self.ray = ray.Ray(self.internal, RAY_MATERIAL2)
        self.ray.set_xyz(0.0, 0.07, 0.65)
        self.ray.endpoint.set_xyz(0.0, 0.3, 0.0)
        self.ray.length = 20
        
  def init(self): pass
  
  def default_mouse_x(self): return 0.0
    
  def set_level(self, level, init = 0):
    level.append(self)
    
    super(Character, self).set_level(level, init)
    
  def begin_action(self, action):
    from py2play.idler import IDLER as root
    
    speed = self.speed
    force = self.force
    
    self.internal.rotate_incline((self.old_speed_x - speed.x) *  6000.0)
    self.perso   .rotate_lateral((self.old_speed_x - speed.x) * 12000.0)
    self.old_speed_x = speed.x
    if speed.z: self.rotate_lateral(-500.0 * speed.x)
    
    self.inercy.convert_to(self)
    speed.set_xyz(
      speed.x * (1.0 - self.inertia) + self.inercy.x * self.inertia,
      speed.y * (1.0 - self.inertia) + self.inercy.y * self.inertia,
      #speed.y,
      speed.z * (1.0 - self.inertia) + self.inercy.z * self.inertia,
      )
    self.inercy.clone(speed)
    self.inercy.convert_to(self.level)
    
    if   action.action == ACTION_ROTUP  : self.turn_vertical( self.rotup_speed)
    elif action.action == ACTION_ROTDOWN: self.turn_vertical(-self.rotup_speed)
    elif action.action == ACTION_ROTLAT:
      self.rotate_lateral(30.0)
    elif action.action == ACTION_DEBUG:
      print self.position(), self.round
        
    #if root.raypick(self.CENTER, DOWN, 1.5, 3, _P, _V):
    #if root.raypick(self.CENTER, DOWN, 0.75, 3, _P, _V):
    #if root.raypick(self.CENTER,  Vector(None, 0.0, -1.0, 0.0), 1.5, 3, _P, _V):
    if root.raypick(self.CENTER, DOWN, self.floating + 0.45, 3, _P, _V):
      _P.convert_to(self.parent)
      self.on_ground = 1
      
      if (self.y < _P.y + self.floating - self.force.y):
        if (not self.scramble) and self.parent:
          self.scramble = min(30.0, ((self.force.y + self.parent.transform_vector(speed.x, speed.y, speed.z, self)[1]) * 7.0) ** 2)
        force.y = 0.0
        self.force.y += _P.y + self.floating - self.y
      elif self.y > _P.y + self.floating + 0.1: force.y -= 0.03
      
      if self.force.y < 0.05:
        if action.action == ACTION_JUMP: force.y += 0.4
        
        elif (action.action != ACTION_ROTUP) and (action.action != ACTION_ROTDOWN):
          _V.convert_to(self)
          
          if _V.y < 0.0: self.turn_incline(10.0)
          elif _V.y < 0.999:
            #axe = _V.cross_product(UP)
            soya.cross_vector(_V, UP, _V2)
            _V2.convert_to(self.parent)
            angle = _V.angle_to(UP)
            self.rotate_axe(-angle / 4.0, _V2.x, _V2.y, _V2.z)
            
    else:
      self.on_ground = 0
      force.y -= 0.03 # comment this line if you want to fly
      if self.y < -35.0: self.hurt(0.2)
      
    force.__imul__(0.95)
    
    if force.y > 0.0: self.turn_vertical(force.y * 14.0)
    
    x, y, z = self.transform_vector(force.x, force.y, force.z, self.parent)
    speed.x += x
    speed.y += y
    speed.z += z
    
    if speed.length() > 0.5: speed.set_length(0.5)
    
    for character in self.level.pushables:
      if not character is self:
        if self.distance_to(character) < self.radius + character.radius:
          _V.set_start_end(self, character)
          
          #_V.y += 0.3
          f = 3.0 * max(0.0, _V.dot_product(self.speed)) * self.weight / character.weight
          if f:
            _V.set_length(f)
            character.push(_V.x, _V.y + 0.3 * f, _V.z)
            f = -0.3 * character.weight / self.weight
            self.push(f * _V.x, f * _V.y, f * _V.z)
#     for character in self.level.characters:
#       if (not character is self) and isinstance(character, Pusher):
#         if self.distance_to(character) < self.radius + character.radius:
#           _V.set_start_end(character, self)
            
#           _V.y += 0.3
#           f = 3.0 * max(0.0, _V.dot_product(character.speed))
#           if f:
#             _V.set_length(f * character.weight / self.weight)
            
#             force.add_vector(_V)
            
    _P2.__init__(self, 0.0, 0.3, 0.0)
    _P2.add_vector(speed)
    _P2.convert_to(root)
    
    context = root.RaypickContext(_P2, max(self.radius, 0.7))
    
    bouree = 0
    
    for vec in (RIGHT, FRONT):
      if context.raypick(_P2, vec, self.radius, 0, _P, _V):
        _P.convert_to(root)
        _V.convert_to(root)
        if vec.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * vec.x, _P2.y - self.radius * vec.y, _P2.z - self.radius * vec.z)
        else:                        _P3.__init__(root, _P2.x + self.radius * vec.x, _P2.y + self.radius * vec.y, _P2.z + self.radius * vec.z)
        
        if force.dot_product(_V) < 0.0: bouree = 1
        
        if (_P.x - _P2.x) * _V.x + (_P.y - _P2.y) * _V.y + (_P.z - _P2.z) * _V.z > 0.0:
          _V.set_start_end(_P2, _P)
        else: _V.set_start_end(_P3, _P)
        
        _P2  .__iadd__(_V)
        speed.__iadd__(_V)
        
    #for _V2 in (Vector(self, 0.0, 0.0, -1.0) % root, Vector(self, 1.0, 0.0, 0.0) % root):
#     if context.raypick(_P2, vec, self.radius, 1, _P, _V):
#       if _V2.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * _V2.x, _P2.y - self.radius * _V2.y, _P2.z - self.radius * _V2.z)
#       else:                       _P3.__init__(root, _P2.x + self.radius * _V2.x, _P2.y + self.radius * _V2.y, _P2.z + self.radius * _V2.z)
      
#       if force.dot_product(_V) < 0.0: bouree = 1
      
#       _V2.set_start_end(_P3, _P)
#       _V.set_length(_V.dot_product(_V2))
#       _P2   .__iadd__(_V)
#       speed.__iadd__(_V)
      
#     _V2.__init__(self, 1.0, 0.0,  0.0); _V2.convert_to(root)
#     if context.raypick(_P2, _V2, self.radius, 1, _P, _V):
#       if _V2.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * _V2.x, _P2.y - self.radius * _V2.y, _P2.z - self.radius * _V2.z)
#       else:                       _P3.__init__(root, _P2.x + self.radius * _V2.x, _P2.y + self.radius * _V2.y, _P2.z + self.radius * _V2.z)
      
#       if force.dot_product(_V) < 0.0: bouree = 1
      
#       _V2.set_start_end(_P3, _P)
#       _V.set_length(_V.dot_product(_V2))
#       _P2   .__iadd__(_V)
#       speed.__iadd__(_V)
      
#     _P2.y += 0.2
#     vec = UP
#     if context.raypick(_P2, vec, self.radius, 1, _P, _V):
#       print "!!!!!!!!!!!!!!!!!!!!!!!"
      
#       if vec.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * vec.x, _P2.y - self.radius * vec.y, _P2.z - self.radius * vec.z)
#       else:
#         if force.y > 0.0: force.y = 0.0 # Stop and break a jump
#         _P3.__init__(root, _P2.x + self.radius * vec.x, _P2.y + self.radius * vec.y, _P2.z + self.radius * vec.z)
        
#       #if (_P.x - _P2.x) * _V.x + (_P.y - _P2.y) * _V.y + (_P.z - _P2.z) * _V.z > 0.0:
#       #  _V.set_start_end(_P2, _P)
#       #else: _V.set_start_end(_P3, _P)
      
#       #_P2   .__iadd__(_V) # Useless since it is the last raypicking.
#       speed.__iadd__(_V)
        
        
    _P2.y += 0.2
    #_V2.__init__(self, 0.0, 1.0, 0.0); _V2.convert_to(root) # DOWN was _V2
    if context.raypick(_P2, DOWN, 0.7, 1, _P, _V):
      _P.convert_to(root)
      _V.convert_to(root)
      if DOWN.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * DOWN.x, _P2.y - self.radius * DOWN.y, _P2.z - self.radius * DOWN.z)
      else:
        if force.y > 0.0: force.y = 0.0 # Stop and break a jump
        _P3.__init__(root, _P2.x + self.radius * DOWN.x, _P2.y + self.radius * DOWN.y, _P2.z + self.radius * DOWN.z)
      if force.dot_product(_V) < 0.0: bouree = 1
      
      _V.set_start_end(_P3, _P)
      
      #_P2   .__iadd__(_V) # Useless since it is the last raypicking.
      speed.__iadd__(_V)
      
    if bouree:
      couic = min(0.5, force.x ** 2 + force.z ** 2)
      if couic > 0.01:
        force.x = force.z = 0.0
        
        if self.hurt(couic) and (len(explosions) < 2): # Limit the number of simultaneous explosions
          explode = Explosion(self.parent)
          explode.move(self)
          
    self.next_pos.clone(speed)
    self.next_pos.convert_to(self.parent)
    self.next_pos.parent = None
    
  def push(self, x, y, z):
    from py2play.idler import IDLER
    IDLER.next_round_tasks.append(lambda : self.force.add_xyz(x, y, z))
    
  def end_round(self):
    # Restore the real position -- advance_time may cause rounding error on the position !!!
    self.move(self.next_pos)
    
  def advance_time(self, proportion):
    soya3d.World.advance_time(self, proportion)
    
    self.add_mul_vector(proportion, self.speed)
    
    if self.scramble:
      if self.scramble <= proportion:
        self.internal.y = 0.0
        self.scramble   = 0.0
      else:
        self.internal.y = abs(math.sin(self.scramble)) * self.scramble * 0.04
        self.scramble  -= proportion / 3.0
      self.perso.y = self.perso_y + self.internal.y
      
  def teleport(self, x, y, z):
    self.set_xyz(x, y, z)
    self.force.set_xyz(0.0, 0.0, 0.0)
    if self.scramble:
      self.scramble   = 0.0
      self.internal.y = 0.0
      self.perso.y    = self.perso_y
    if hasattr(self, "ray"): self.ray.zap()
    
  def hurt(self, life):
    """Kills LIFE life to this character. Returns true if the character is still alive."""
    self.life -= life
    if self.life < 0.0:
      sound.play("explose-2.wav", self)
      self.die()
    else:
      sound.play("explose-1.wav", self)
      return 1
    
  def die(self): 
    explode = Explosion(self.parent)
    explode.move(self)
    explode.life = 60
    explode.particle_width = explode.particle_height = 0.7
    self.level.remove_character(self)


class NonPlayerCharacter(Character, py2play_character.NonPlayerCharacter):
  clan = 0
  
  def __init__(self, parent = None):
    py2play_character.NonPlayerCharacter.__init__(self)
    Character.__init__(self, parent)
    
    import slune.controler
    self.controler = slune.controler.StackControler([self.IAControler()])
    
  def choose_action(self):
    return self.controler.next()
    
  def begin_round(self):
    py2play_character.NonPlayerCharacter.begin_round(self)
    
  def begin_action(self, action):
    self.speed.set_xyz(action.mouse_x / 120.0 * self.maxspeedx, 0.0, (action.mouse_y * 0.2 - 0.28) * self.maxspeedz)
    
    Character.begin_action(self, self.current_action)
    
  def __getstate__(self):
    state = Character.__getstate__(self)
    del state["controler"]
    return state
  def __setstate__(self, state):
    state = Character.__setstate__(self, state)
    import slune.controler
    self.controler = slune.controler.StackControler([self.IAControler()])
    
  def IAControler(character):
    while 1: yield Action(ACTION_WAIT, 0.0, 1.4) # Does nothing !
    
  
class Pushable(soya3d.Volume, Pusher):
  clan = 0
  
  def __init__(self, parent = None):
    soya3d.Volume.__init__(self, parent)
    
    self.radius      = 1.0
    self.radius_y    = 1.0
    self.weight      = 1.0
    self.life        = 1.0
    self.on_ground   = 1 # 0 ???
    
    self.speed       = Vector(self)
    self.force       = Vector(None)
    self.solid       = 0
    self.worth_playing = 0
    
    self.CENTER = Point (self, 0.0, 0.3,  0.0)
    
    self.next_pos = Point(self, 0.0, 0.0, 0.0)
    
  def begin_round(self):
    force = self.force
    if not self.worth_playing:
      if (abs(force.x) > 0.1) or (abs(force.y) > 0.1) or (abs(force.z) > 0.1):
        self.worth_playing = 1
      else: return
      
    speed = self.speed
    speed.set_xyz(0.0, 0.0, 0.0)
    
    from py2play.idler import IDLER as root
    
    if root.raypick(self.CENTER, DOWN, self.radius_y / 2.0 + 0.45, 3, _P, _V):
      _P.convert_to(self.parent)
      self.on_ground = 1
      
      if (self.y < _P.y + self.radius_y / 2.0 - self.force.y):
        force.y = 0.0
        self.force.y += _P.y + self.radius_y / 2.0 - self.y
      elif self.y > _P.y + self.radius_y / 2.0 + 0.1: force.y -= 0.03
      
      if self.force.y < 0.05:
        _V.convert_to(self)
        
        if _V.y < 0.0: self.turn_incline(10.0)
        elif _V.y < 0.999:
          #axe = _V.cross_product(UP)
          soya.cross_vector(_V, UP, _V2)
          _V2.convert_to(self.parent)
          angle = _V.angle_to(UP)
          self.rotate_axe(-angle / 4.0, _V2.x, _V2.y, _V2.z)
        else:
          self.worth_playing = 0
          
    else:
      self.on_ground = 0
      force.y -= 0.03 # comment this line if you want to fly
      if self.y < -35.0: self.die()
      
    force.__imul__(0.95)
    
    if force.y > 0.0: self.turn_vertical(force.y * 14.0)
    
    x, y, z = self.transform_vector(force.x, force.y, force.z, self.parent)
    speed.x += x
    speed.y += y
    speed.z += z
    
    if speed.length() > 0.5: speed.set_length(0.5)
    
    for character in self.parent.pushables:
      if not character is self:
        if self.distance_to(character) < self.radius + character.radius:
          _V.set_start_end(self, character)
          
          _V.y += 0.3
          f = 2.5 * max(0.0, _V.dot_product(self.speed))
          if f:
            _V.set_length(f * self.weight / character.weight)
            character.push(_V.x, _V.y, _V.z)
            
    _P2.__init__(self, 0.0, self.radius_y / 2.0, 0.0)
    _P2.add_vector(speed)
    _P2.convert_to(root)
    
    context = root.RaypickContext(_P2, max(self.radius, 0.7))
    
    bouree = 0
    
    for vec in (RIGHT, FRONT):
      if context.raypick(_P2, vec, self.radius, 0, _P, _V):
        _P.convert_to(root)
        _V.convert_to(root)
        if vec.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * vec.x, _P2.y - self.radius * vec.y, _P2.z - self.radius * vec.z)
        else:                        _P3.__init__(root, _P2.x + self.radius * vec.x, _P2.y + self.radius * vec.y, _P2.z + self.radius * vec.z)
        
        if force.dot_product(_V) < 0.0: bouree = 1
        
        if (_P.x - _P2.x) * _V.x + (_P.y - _P2.y) * _V.y + (_P.z - _P2.z) * _V.z > 0.0:
          _V.set_start_end(_P2, _P)
        else: _V.set_start_end(_P3, _P)
        
        _P2  .__iadd__(_V)
        speed.__iadd__(_V)
        
    _P2.y += 0.2
    if context.raypick(_P2, DOWN, self.radius_y, 1, _P, _V):
      _P.convert_to(root)
      _V.convert_to(root)
      if DOWN.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius_y * DOWN.x, _P2.y - self.radius_y * DOWN.y, _P2.z - self.radius_y * DOWN.z)
      else:
        if force.y > 0.0: force.y = 0.0 # Stop and break a jump
        _P3.__init__(root, _P2.x + self.radius_y * DOWN.x, _P2.y + self.radius_y * DOWN.y, _P2.z + self.radius_y * DOWN.z)
      if force.dot_product(_V) < 0.0: bouree = 1
      
      _V.set_start_end(_P3, _P)
      
      #_P2   .__iadd__(_V) # Useless since it is the last raypicking.
      speed.__iadd__(_V)
      
    if bouree:
      couic = min(0.5, force.x ** 2 + force.z ** 2)
      if couic > 0.01:
        force.x = force.z = 0.0
        
        if self.hurt(couic) and (len(explosions) < 2): # Limit the number of simultaneous explosions
          explode = Explosion(self.parent, nb_particles = 15)
          explode.move(self)
          
    self.next_pos.clone(speed)
    self.next_pos.convert_to(self.parent)
    self.next_pos.parent = None
    
  def push(self, x, y, z):
    from py2play.idler import IDLER
    IDLER.next_round_tasks.append(lambda : self.force.add_xyz(x, y, z))
    
  def end_round(self):
    if self.worth_playing:
      # Restore the real position -- advance_time may cause rounding error on the position !!!
      self.move(self.next_pos)
      
  def advance_time(self, proportion):
    if self.worth_playing:
      self.add_mul_vector(proportion, self.speed)
      
  def hurt(self, life):
    """Kills LIFE life to this character. Returns true if the character is still alive."""
    sound.play("explose-1.wav", self)
    
    self.life -= life
    if self.life < 0.0: self.die()
    else: return 1
    
  def die(self): 
    explode = Explosion(self.parent, nb_particles = 20)
    explode.move(self)
    explode.life = 30
    explode.particle_width = explode.particle_height = 0.7
    self.parent.pushables.remove(self)
    self.parent.remove(self)
    
def Barrier(parent):
  p = Pushable(parent)
  p.set_shape(model.Shape.get("gate3"))
  p.radius_y = 0.5
  p.life     = 0.5
  p.weight   = 0.6
  return p
    
class Remorque(NonPlayerCharacter, py2play_character.NonPlayerCharacter):
  clan = 0
  
  def __init__(self, parent = None, remorqued_by = None, remorqued_distance = 1.5):
    py2play_character.NonPlayerCharacter.__init__(self)
    NonPlayerCharacter.__init__(self, parent, 0)
    
    self.remorqued_by       = remorqued_by
    self.remorqued_distance = remorqued_distance
    
    self.set_vehicle(1)
    
  def begin_action(self, action):
    self.current_action = Action()
    self.speed.set_xyz(0.0, 0.0, 0.0)
    
    if self.remorqued_by:
      _V.set_start_end(self, self.remorqued_by)
      _V.convert_to(self)
      _V.set_length(_V.length() - self.remorqued_distance)
      if _V.length() > 0.48: _V.set_length(0.48)
      _V.y = _V.y - self.force.y
      self.speed.set_xyz(_V.x / 20.0, _V.y, _V.z)
    else:
      self.speed.set_xyz(0.0, 0.0, 0.0)
      
    Character.begin_action(self, self.current_action)
    
RAY_MATERIAL = model.Material()
RAY_MATERIAL.diffuse = (0.7, 0.4, 0.95, 1.0)
RAY_MATERIAL.additive_blending = 1

RAY_MATERIAL2 = model.Material()
RAY_MATERIAL2.diffuse = (0.9, 0.6, 0.3, 1.0)
RAY_MATERIAL2.additive_blending = 1

def PlayerLight(parent):
  light = soya3d.Light(parent)
  light.set_xyz(0.0, 1.2, -0.5)
  light.diffuse   = (0.9, 0.8, 1.0, 1.0)
  light.specular  = (0.0, 0.0, 0.0, 1.0)
  light.quadratic = 0.005
  light.linear    = 0.2
  light.constant  = 0.0
  return light

class PlayerCharacter(Character, py2play_character.PlayerCharacter, Competitor):
  clan = 1
  
  def __init__(self, player = None):
    py2play_character.PlayerCharacter.__init__(self, player)
    Character.__init__(self)
    
    if player: player.default_action = action.ACTION_WAIT

    self.set_perso  (globdef.CHARACTER.lower())
    self.set_vehicle(globdef.VEHICLE)
    
    self.light = PlayerLight(self)
    
    if globdef.VERIFICATION_SERVER and self.player: self.logon_verif()
    
  def displayname(self): return self.perso_name.title() + " (" + self.player.name + ")"
  
  def default_mouse_x(self):
    return self.player.controler.mouse_x
  
  def logon_verif(self):
    """Log on the verif server."""
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((globdef.VERIFICATION_SERVER, 36081))
    self.player.verif = s.makefile("w", 0)
    self.player.verif.write("P") # Check position
    
  def begin_round(self):
    import soya.math3d
    
    if self.player.active: self.plan_action(self.player.controler.next())
    else:                  self.controler.next()
    
    py2play_character.PlayerCharacter.begin_round(self)
    Character.begin_round(self) # XXX Actually World.begin_round; not needed... ?
    
  def begin_action(self, action):
    if globdef.VERIFICATION_SERVER:
      import cPickle as pickle
      pickle.dump((self.player.name, self.round, (self.x, self.y, self.z), None), self.player.verif)
      
    # Compute the speed
    self.speed.set_xyz(action.mouse_x / 120.0 * self.maxspeedx, 0.0, (action.mouse_y * 0.2 - 0.28)  * self.maxspeedz)
    
    if action.action == ACTION_STARTGAME:
      from py2play.idler import IDLER
      if IDLER.waiting_until_ready: IDLER.start_game()
      else:                         IDLER.unpause()
      
    Character.begin_action(self, action)
    
  def get_controler(self):
    return self.player.controler
    if self.player.active: return self.player.controler
    return None
  controler = property(get_controler)
  
  def hurt(self, life):
    if   self.level.difficulty == 0: return Character.hurt(self, 0.6 * life)
    elif self.level.difficulty == 1: return Character.hurt(self, 0.9 * life)
    else:                            return Character.hurt(self, life)
    
  def die(self):
    if len(explosions) < 2: # Limit the number of simultaneous explosions
      explode = Explosion(self.parent)
      explode.move(self)
      explode.life = 60
      explode.particle_width = explode.particle_height = 0.7
    
    self.remove(self.light)
    self.level.add(self.light)
    self.light.move(self)
    
    oldlevel = self.level
    self.level.remove_character(self)
    self.player.level = oldlevel # May be required !!!
    
    if self.player.active:
      from py2play.idler import IDLER
      import soya.game.camera as game_camera
      IDLER.show_blackbands()
      IDLER.camera.add_traveling(game_camera.FixTraveling(IDLER.camera, Vector(IDLER.camera, 0.0, 0.0, -1.0)))
      IDLER.message(_("__gameover__"), 100000)
      
      # As the character has been removed from the level, the controler
      # does no longer listen for event such as the "escape" key.
      AllowQuit(IDLER)
      
  def __getstate__(self):
    return self._treat_getstate(Character.__getstate__(self))
  
  def __setstate__(self, state):
    Character.__setstate__(self, state)
    self._treat_setstate(state)
    
    # Loaded => imply this character is a phantom one !!!
    from slune.controler import PhantomControler
    self.player.controler = PhantomControler()
    
    if globdef.VERIFICATION_SERVER: self.logon_verif()
    
  def __repr__(self):
    if self.player: return "<Vehicle %s@%s:%s>" % (self.player.name, self.player.host, self.player.port)
    

explosions = []
class Explosion(particle.SpriteSystem):
  def __init__(self, parent = None, material = None, nb_particles = 30):
    particle.SpriteSystem.__init__(self, parent, particle.FireGenerator(initial_speed = 0.4), material or particle._default(), nb_particles = nb_particles)
    self.particle_width = self.particle_height = 0.5
    self.life           = 25
    explosions.append(self)
    
  def begin_round(self):
    if   self.life >  0: self.life -= 1
    elif self.life == 0:
      self.generator = None
      self.removable = 1
      explosions.remove(self)
      self.life = -1
      
class LifeBar(widget.Widget):
  def __init__(self, character = None):
    self.character = character
    
    self.left   = soya.get_screen_width() - 30.0
    self.top    = 10.0
    self.width  = 20.0
    self.height = soya.get_screen_height() - 20.0
    
  def render(self):
    if self.character:
      import _soya
      _soya.glBegin(_soya.GL_QUADS)
      _soya.glColor4f(0.7, 0.0, 0.0, 1.0)
      
      _soya.glVertex2f(self.left             , self.top + self.height)
      _soya.glVertex2f(self.left + self.width, self.top + self.height)
      
      top = self.top + (1.0 - self.character.life) * self.height
      _soya.glColor4f(1.0, 0.7, 0.0, 1.0)
      _soya.glVertex2f(self.left + self.width, top)
      _soya.glVertex2f(self.left             , top)
      _soya.glEnd()


class AllowQuit(soya3d.Volume):
  def begin_round(self):
    for event in soya.process_event():
      if (event[0] == soya.MOUSEBUTTONDOWN) or (event[0] == soya.KEYDOWN):
        from py2play.idler import IDLER
        IDLER.end_game()
        

